/* -*- tab-width:4;c-file-style:"cc-mode"; -*- */
/*
 * libtheora_utils.c - Encoder module that uses libtheora
 * Copyright (C) 2008  BBC, Anuradha Suraparaju <asuraparaju@users.sourceforge.dot.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ctx program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#ifdef HAVE_THEORA
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "ogg/ogg.h"
#include "theora/theora.h"
#include "libavutil/common.h"
#include "libavformat/avformat.h"
#include "oggmuxutils.h"
#include "ffmpeg2dirac.h"
#include "ff2dirac_theora.h"

static int ff2libtheora_encode_header(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    theora_encode_header(dd->di, op);
    return 0;
}

static int ff2libtheora_encode_clear (ff2dirac_video_enc_state_t *dd)
{
    ff2theora_privdata_t *info = dd->vcodec->priv_data;
    theora_comment_clear(&info->tc);
    theora_info_clear(&info->ti);
    theora_clear(dd->di);
    av_free(dd->di);
    dd->di = NULL;
    return 0;
}

static int ff2libtheora_load_frame(ff2dirac_video_enc_state_t *dd, AVPicture *pic)
{
    theora_state *td = dd->di;
    ff2theora_privdata_t *info= dd->vcodec->priv_data;
    yuv_buffer yuv;

    yuv.y_width = info->ti.width;
    yuv.y_height = info->ti.height;
    yuv.y_stride = pic->linesize[0];

    yuv.uv_width = info->ti.width / 2;
    yuv.uv_height = info->ti.height / 2;
    yuv.uv_stride = pic->linesize[1];

    yuv.y = pic->data[0];
    yuv.u = pic->data[1];
    yuv.v = pic->data[2];

    theora_encode_YUVin(td, &yuv);
    return 0;
}

static int ff2libtheora_encode_packetout(ff2dirac_video_enc_state_t *dd, int e_o_s, ogg_packet *op)
{
    return theora_encode_packetout(dd->di, e_o_s, op);
}

static const char *metadata_tag[] = {
    "ARTIST",
    "TITLE",
    "DATE",
    "LOCATION",
    "ORGANIZATION",
    "COPYRIGHT",
    "LICENSE",
    "CONTACT",
    "SOURCE HASH",
};

static int ff2libtheora_encode_init (ff2dirac ctx, ff2dirac_video_enc_state_t *dd)
{
    int i;
    int csize = sizeof(ctx->metadata)/sizeof(ctx->metadata[0]);
    ff2theora_privdata_t *info = av_mallocz(sizeof(ff2theora_privdata_t));
    theora_state *td = av_mallocz(sizeof(theora_state));

    theora_info_init (&info->ti);
    info->ti.width = ctx->frame_width;
    info->ti.height = ctx->frame_height;
    info->ti.frame_width = ctx->picture_width;
    info->ti.frame_height = ctx->picture_height;
    info->ti.offset_x = ctx->frame_x_offset;
    info->ti.offset_y = ctx->frame_y_offset;
    info->ti.fps_numerator = ctx->framerate.num;
    info->ti.fps_denominator = ctx->framerate.den;
    /* ctx is pixel aspect ratio */
    info->ti.aspect_numerator=ctx->aspect_numerator;
    info->ti.aspect_denominator=ctx->aspect_denominator;
    // FIXME: is all input material with fps==25 OC_CS_ITU_REC_470BG?
    // guess not, commandline option to select colorspace would be the best.
    if((ctx->fps-25)<1)
        info->ti.colorspace = OC_CS_ITU_REC_470BG;
    else if(abs(ctx->fps-30)<1)
        info->ti.colorspace = OC_CS_ITU_REC_470M;
    else
        info->ti.colorspace = OC_CS_UNSPECIFIED;

    info->ti.target_bitrate = ctx->video_bitrate;
    info->ti.quality = rint(ctx->video_quality/10*6.3);
    if (info->ti.quality < 0)
        info->ti.quality = 0;
    if (info->ti.quality > 63)
        info->ti.quality = 63;
    info->ti.dropframes_p = 0;
    info->ti.keyframe_auto_p = 1;
    info->ti.keyframe_frequency = ctx->keyint;
    info->ti.keyframe_frequency_force = ctx->keyint;
    info->ti.keyframe_data_target_bitrate = info->ti.target_bitrate * 5;
    info->ti.keyframe_auto_threshold = 80;
    info->ti.keyframe_mindistance = 8;
    info->ti.noise_sensitivity = 1;
    // range 0-2, 0 sharp, 2 less sharp,less bandwidth
    info->ti.sharpness = ctx->sharpness;

    theora_comment_init(&info->tc);
    for (i = 0; i < csize; ++i) {
        if (ctx->metadata[i])
            theora_comment_add_tag(&info->tc, metadata_tag[i], ctx->metadata[i]);
    }

    theora_comment_add_tag(&info->tc, "ENCODER", PACKAGE_STRING);

    theora_encode_init(td, &info->ti);
    dd->di = td;
    if (ctx->speed_level >= 0)
    {
        int max_speed_level;

        theora_control(td, TH_ENCCTL_GET_SPLEVEL_MAX, &max_speed_level, sizeof(int));
        if(ctx->speed_level > max_speed_level)
          ctx->speed_level = max_speed_level;
        theora_control(td, TH_ENCCTL_SET_SPLEVEL, &ctx->speed_level, sizeof(int));
    }
    dd->vcodec->priv_data = info;
    return 0;
}

extern double ff2libtheora_granule_time(ff2dirac_video_enc_state_t *dd, ogg_int64_t granulepos)
{
    return theora_granule_time (dd->di, granulepos);
}

int ff2libtheora_granule_shift(ff2dirac_video_enc_state_t *dd)
{
    theora_info *ti = &((ff2theora_privdata_t *)dd->vcodec->priv_data)->ti;
    return theora_granule_shift(ti);
}

int ff2libtheora_encode_comment(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    theora_comment *tc = &((ff2theora_privdata_t *)dd->vcodec->priv_data)->tc;
    theora_encode_comment(tc, op);
    return 0;
}

int ff2libtheora_encode_tables(ff2dirac_video_enc_state_t *dd, ogg_packet *op)
{
    theora_encode_tables(dd->di, op);
    return 0;
}

ff2dirac_vcodec_t ff2libtheora_encoder = {
  VCODEC_THEORA,
  NULL,
  ff2libtheora_encode_init,
  ff2libtheora_encode_header,
  ff2libtheora_load_frame,
  ff2libtheora_encode_packetout,
  ff2libtheora_encode_clear,
  ff2libtheora_granule_time,
  ff2libtheora_granule_shift,
  ff2libtheora_encode_comment,
  ff2libtheora_encode_tables
};
#endif
